//
//  PDGameRoot.cpp
//  PolyDemo
//
//  This is the root class for a single game scene.  It shows how to subclass RootLayer to
//  make a new game.  This is a standard design pattern that we have used in all our samples.
//  You may find that you want more than one scene in your game, but that is not always
//  necessary.  You may find it easier to work with a single scene.
//
//  Author: Walker White
//  Version: 2/10/16
//
#include "PDGameRoot.h"

using namespace cocos2d;


#pragma mark -
#pragma mark Loading Screen

#define DEFAULT_FONT_SIZE   64.0f
#define LOADING_FONT_NAME   "felt"
#define LOADING_MESSAGE     "Loading..."

#define COCOS_TEXTURE   "cocos"
#define EARTH1_TEXTURE  "earth1"
#define EARTH2_TEXTURE  "earth2"

#define MAX_SEGMENTS    12
#define POLY_RADIUS     160


#pragma mark -
#pragma mark Gameplay Control

/**
 * Starts the layer, allocating initial resources
 *
 * This method is used to bootstrap the game.  It should start up an asset manager
 * and load initial assets.
 */
void PolygonRoot::start() {
    int scene = AssetManager::getInstance()->createScene();
    
    FontLoader* fonts = FontLoader::create();
    fonts->setDefaultSize(DEFAULT_FONT_SIZE);
    AssetManager::getInstance()->at(scene)->attach<TTFont>(fonts);
    AssetManager::getInstance()->at(scene)->attach<Texture2D>(TextureLoader::create());
    AssetManager::getInstance()->startScene(scene);
    
    // Create a "loading" screen
    _preloaded = false;
    _activated = false;
    displayLoader();
    RootLayer::start(); // YOU MUST END with call to parent
    
    // Activate the input controller
    _input.init();
    _input.start();
    
    // Set the bounds
    Size size = getContentSize();
    Vec2 center(size.width/2.0f,size.height/2.0f);
    _bounds.size   = size;
    _bounds.origin.x = -size.width;
    _bounds.origin.y = -size.height;
    
    // Give us some room to scroll
    setContentSize(size*2);
    setAnchorPoint(Vec2(0.5,0.5));
    setPosition(-size.width/2.0f,-size.height/2.0f);
}

/**
 * Stops the layer, releasing all resources
 *
 * This method is used to clean-up any allocation that occurred in either start or
 * update.  While Cocos2d does have rudimentary garbage collection, you still have
 * to release any objects that you have retained.
 */
void PolygonRoot::stop() {
    RootLayer::stop();  // YOU MUST BEGIN with call to parent
    int scene = AssetManager::getInstance()->getCurrentIndex();
    _input.stop();

    AssetManager::getInstance()->stopScene(scene);
}

/**
 * Updates the game for a single animation frame
 *
 * This method is called every animation frame.  There is no draw() or render()
 * counterpoint to this method; drawing is done automatically in the scene graph.
 * However, this method is responsible for updating any transforms in the scene graph.
 *
 * @param  dt   the time in seconds since last update
 */
void PolygonRoot::update(float deltaTime) {
    RootLayer::update(deltaTime);  // YOU MUST BEGIN with call to parent
    
    if (_activated) {
        _input.update(deltaTime);
        panScreen(_input.getPanDelta());
    } else if (!_preloaded) {
        _preloaded = AssetManager::getInstance()->getCurrent()->isComplete();
    } else {
        displayPolygons();
        _activated = true;
    }
}


#pragma mark -
#pragma mark Internal Helpers

/**
 * Builds the scene graph for the loading screen.
 *
 * The loading screen is what we run while we are waiting for the asynchronous
 * loader for finish with the textures.  Right now, the loading screen is just
 * the word "Loading...".  We do this because the loading screen can only use
 * assets that have been loaded already, and the font is the only thing that
 * is guaranteed to be loaded at start.
 */
void PolygonRoot::displayLoader() {
    // Load the font NOW
    AssetManager::getInstance()->getCurrent()->load<TTFont>(LOADING_FONT_NAME, "fonts/MarkerFelt.ttf");
    
    Size size = getContentSize();
    Vec2 center(size.width/2.0f,size.height/2.0f);

    // Create the message label.
    auto label = Label::create();
    label->setTTFConfig(AssetManager::getInstance()->getCurrent()->get<TTFont>(LOADING_FONT_NAME)->getTTF());
    label->setAnchorPoint(Vec2(0.5f,0.5f));
    label->setPosition(center);
    label->setString(LOADING_MESSAGE);
    
    // Add the label as a child to this layer
    addChild(label, 1);
    
    // Load the textures (Autorelease objects)
    Texture2D::TexParams params;
    params.wrapS = GL_REPEAT;
    params.wrapT = GL_REPEAT;
    params.magFilter = GL_LINEAR;
    params.minFilter = GL_LINEAR;
    
    SceneManager* _assets = AssetManager::getInstance()->getCurrent();
    TextureLoader* tloader = (TextureLoader*)_assets->access<Texture2D>();
    tloader->loadAsync(COCOS_TEXTURE, "textures/cocos2d.png");
    tloader->loadAsync(EARTH1_TEXTURE, "textures/earthtile1.png", params);  // Earth texture is tiled
    tloader->loadAsync(EARTH2_TEXTURE, "textures/earthtile2.png");          // Earth texture is not tiled

}

/**
 * Builds the scene graph for the game itself
 *
 * Normally we offload this to a game controller that corresponds to a single
 * level in the game.
 */
void PolygonRoot::displayPolygons() {
    // Clear what was there before
    removeAllChildren();

    // We need to know the content scale for resolution independence
    // If the device is higher resolution than 1024x576, Cocos2d will scale it
    // THIS DOES NOT FIX ASPECT RATIO PROBLEMS
    // If you are using a device with a 3:2 aspect ratio, you will need to
    // completely redo the level layout.  We can help if this is an issue.
    float cscale = Director::getInstance()->getContentScaleFactor();
    
    Size size = getContentSize();
    Vec2 center(size.width/2.0f,size.height/2.0f);

    SceneManager* _assets = AssetManager::getInstance()->getCurrent();

    // Put the instructions in the middle of the screen
    auto label = Label::create();
    label->setTTFConfig(_assets->get<TTFont>(LOADING_FONT_NAME)->getTTF());
    label->setString("Pan (two fingers) to scroll.");
    label->setPosition(center);
    addChild(label, 0);
    
    // Create a base polygon (circle) for use in all the example
    float vertices[2*MAX_SEGMENTS];
    for(int ii = 0; ii < MAX_SEGMENTS; ii++) {
        vertices[2*ii  ] = (1+cos((ii*M_PI*2)/MAX_SEGMENTS))*(POLY_RADIUS);
        vertices[2*ii+1] = (1+sin((ii*M_PI*2)/MAX_SEGMENTS))*(POLY_RADIUS);
    }
    Poly2 poly(&(vertices[0]),MAX_SEGMENTS*2,0);
    
    // To properly scale to a texture, you need to adjust by the design scale
    poly /= cscale;
    
    // Get the three textures
    Texture2D* cocos = _assets->get<Texture2D>(COCOS_TEXTURE);
    Texture2D* earth1 = _assets->get<Texture2D>(EARTH1_TEXTURE);
    Texture2D* earth2 = _assets->get<Texture2D>(EARTH2_TEXTURE);

    // Call this on a polygon to get a solid shape
    poly.triangulate();  // DELETE ME AND SEE WHAT HAPPENS

    // Create the top left sprite
    auto topleft = PolygonNode::createWithTexture(cocos,poly);
    topleft->setPosition(center.x-size.width/4.0f, center.y+size.height/4.0f);
    topleft->setScale(cscale);
    addChild(topleft, 1);

    // Create the top middle sprite (tiled texture)
    auto topmid = PolygonNode::createWithTexture(earth1,poly);
    topmid->setPosition(center.x, center.y+size.height/4.0f);
    topmid->setScale(cscale);
    addChild(topmid, 1);

    // Create the top right sprite (clamped texture)
    auto topright = PolygonNode::createWithTexture(earth2,poly);
    topright->setPosition(center.x+size.width/4.0f, center.y+size.height/4.0f);
    topright->setScale(cscale);
    addChild(topright, 1);
    
    // Now we play with paths and wireframes
    
    // Create the bottom left wireframe
    poly.traverse(Poly2::Traversal::INTERIOR); // DELETE ME AND SEE WHAT HAPPENS
    
    auto botleft = WireNode::createWithPoly(poly);
    botleft->setColor(Color3B::YELLOW);
    botleft->setPosition(center.x-size.width/4.0f, center.y-size.height/4.0f);
    botleft->setScale(cscale);
    addChild(botleft, 1);
    
    // Create the bottom middle path (closed and bevel joints)
    auto botmid = PathNode::createWithPoly(poly, 10, true);
    botmid->setJoint(Poly2::Joint::BEVEL);
    botmid->setTexture(earth1);
    botmid->setPosition(center.x, center.y-size.height/4.0f);
    botmid->setScale(cscale);
    addChild(botmid, 1);

    // Create the bottom middle path (open with no joints and round cap)
    auto botright = PathNode::createWithPoly(poly, 10, false);
    botright->setCap(Poly2::Cap::ROUND);
    botright->setTexture(earth1);
    botright->setPosition(center.x+size.width/4.0f, center.y-size.height/4.0f);
    botright->setScale(cscale);
    addChild(botright, 1);
    
}

/**
 * Pans the screen by the amount given by the input
 *
 * @param delta The pan offset
 */
void PolygonRoot::panScreen(const Vec2& delta) {
    if (delta.lengthSquared() == 0) {
        return;
    }
    
    float x = getPosition().x+delta.x;
    float y = getPosition().y+delta.y;
    if (!_bounds.containsPoint(Vec2(x,y))) {
        x = MAX(x,_bounds.origin.x);
        x = MIN(x,_bounds.origin.x+_bounds.size.width);
        y = MAX(y,_bounds.origin.y);
        y = MIN(y,_bounds.origin.y+_bounds.size.height);
    }
    setPosition(x,y);
}



