//
//  PDInputController.cpp
//  PolyDemo
//
//  This input controller is primarily designed for touch and mouse control.  On mobile
//  you will notice that we need to use two fingers to pan.  However, when a mouse is
//  available, we can pan without that overhead.
//
//  Author: Walker White
//  Version: 2/10/16
//
#include "PDInputController.h"

/** Square of the two finger tolerance (found by experimentation) */
#define PAN_DELTA 50000.0f

#pragma mark -
#pragma mark Input Controller
/**
 * Creates a new input controller.
 *
 * This constructor does NOT do any initialzation.  It simply allocates the
 * object. This makes it safe to use this class without a pointer.
 */
InputController::InputController() :
_active(false),
_touchListener(nullptr),
_mouseListener(nullptr)
{
    _mousepan = false;
    _panid1 = -1;
    _panid2 = -1;
}


/**
 * Disposes of this input controller, releasing all listeners.
 */
InputController::~InputController() {
    if (_touchListener != nullptr) {
        _touchListener->release();
        _touchListener = nullptr;
    }
    if (_mouseListener != nullptr) {
        _mouseListener->release();
        _mouseListener = nullptr;
    }
}

/**
 * Initializes the input controller
 *
 * This method works like a proper constructor, initializing the input
 * controller and allocating memory.  However, it still does not activate
 * the listeners.  You must call start() do that.
 *
 * @return true if the controller was initialized successfully
 */
bool InputController::init() {
    _timestamp = current_time();

    // Create the touch listener. This is an autorelease object.
    _touchListener = TouchListener::create();
    if (_touchListener == nullptr) {
        return false;
    }

    // Create the mouse listener. This is an autorelease object.
    _mouseListener = MouseListener::create();
    if (_mouseListener == nullptr) {
        return false;
    }

    // Hook up touch callbacks
    _touchListener->retain();
    _touchListener->onTouchBegan = [this](Touch* t, timestamp_t time) {
        return touchBeganCB(t,time);
    };
    _touchListener->onTouchMoved = [=](Touch* t, timestamp_t time) {
        return this->touchMovedCB(t,time);
    };
    _touchListener->onTouchEnded = [=](Touch* t, timestamp_t time) {
        return this->touchEndedCB(t,time);
    };
    _touchListener->onTouchCancelled = [=](Touch* t, timestamp_t time) {
        return this->touchCancelCB(t,time);
    };
    
    // Hook up mouse callbacks
    _mouseListener->retain();
    _mouseListener->onMouseDown = [this](EventMouse* event, timestamp_t time) {
        return mouseDownCB(event,time);
    };
    _mouseListener->onMouseMove = [=](EventMouse* event, timestamp_t time) {
        return this->mouseMoveCB(event,time);
    };
    _mouseListener->onMouseUp = [=](EventMouse* event, timestamp_t time) {
        return this->mouseUpCB(event,time);
    };
    
    return true;
}

/**
 * Starts the input processing for this input controller.
 *
 * This method must be called AFTER the input controller is initialized
 */
void InputController::start() {
    if (!_active) {
        _active = true;
        // Priority the input
        // BUG: Cocos2D always prioritizes touch
        _mouseListener->start(1);
        _touchListener->start(2);
    }
}

/**
 * Stops the input processing for this input controller.
 *
 * This method will not dispose of the input controller. It can be restarted.
 */
void InputController::stop() {
    if (_active) {
        _active = false;
        _touchListener->stop();
        _mouseListener->stop();
    }
}

/**
 * Processes the currently cached inputs.
 *
 * This method is used to to poll the current input state.  It will gather the
 * delta difference in the touches. Depending on the OS, we may see multiple updates 
 * of the same touch in a single animation frame, so we need to accumulate all of 
 * the data together.
 */
void InputController::update(float dt) {
    if (!_active) {
        return;
    }
    
    // If pan is active, compute the delta distance.
    if (_panid1 != -1 || _mousepan) {
        _pandelta = current[_panid1]-previous[_panid1];
    } else if (_mousepan) {
        _pandelta = current[_panid1]-previous[_panid1];
    } else {
        _pandelta.x = 0.0f;
        _pandelta.y = 0.0f;
    }
    
    // Update previous from current
    for (std::unordered_map<int,Vec2>::iterator it = current.begin() ;
         it != current.end(); ++it) {
        previous[it->first] = it->second;
    }
}


#pragma mark -
#pragma mark Touch Panning

/**
 * Callback for the beginning of a touch event
 *
 * @param t     The touch information
 * @param event The associated event
 *
 * @return True if the touch was processed; false otherwise.
 */
 bool InputController::touchBeganCB(Touch* t, timestamp_t time) {
     if (_mousepan) {
         return false;
     }
     
     // Track the touch location by ID
     previous[t->getID()] = t->getLocation();
     current[t->getID()] = t->getLocation();
 
     // Search for a two finger pan if there is none.
     if (_panid1 == -1 && previous.size() > 1) {
         reacquirePan();
     }
     return true;
 }
 
 /**
  * Callback for the end of a touch event
  *
  * @param t     The touch information
  * @param event The associated event
  */
 void InputController::touchEndedCB(Touch* t, timestamp_t time) {
    if (_mousepan) {
        return;
    }
 
    // Remove the touch from tracking.
    previous.erase(t->getID());
    current.erase(t->getID());
 
    // If it was part of the two finger pan, then stop.
    if (_panid1 == t->getID() || _panid2 == t->getID()) {
        _panid1 = -1; _panid2 = -1;
        // Reacquire if possible
        reacquirePan();
    }
 }
 
 
 /**
  * Callback for a touch movement event
  *
  * @param t     The touch information
  * @param event The associated event
  */
 void InputController::touchMovedCB(Touch* t, timestamp_t time) {
     if (_mousepan) {
         return;
     }

     // Put the last movement in current.
     current[t->getID()] = t->getLocation();
 }
 
 /**
  * Callback for the cancellation of a touch event
  *
  * Cancellation occurs when an external event—for example,
  * an incoming phone call—disrupts the current app’s event
  * processing.
  *
  * @param t     The touch information
  * @param event The associated event
  */
 void InputController::touchCancelCB(Touch* t, timestamp_t time) {
     // Abort all panning
     previous.clear();
     current.clear();
     _panid1 = -1;
     _panid2 = -1;
     _mousepan = false;
 }

/**
 * Search for a new pan
 *
 * This method searches for two fingers that are close together. The pan will
 * map to the two touch ids.  The pan will last as long as both fingers are
 * down.  The fingers can be moved apart and the pan will still continue.
 * Only the first touch will determine the pan position.
 */
void InputController::reacquirePan() {
    for (std::unordered_map<int,Vec2>::iterator it = previous.begin() ;
         it != previous.end(); ++it) {
        for (std::unordered_map<int,Vec2>::iterator jt = previous.begin() ;
             jt != previous.end(); ++jt) {
            if (jt->first > it->first) {
                Vec2 diff = it->second-jt->second;
                if (diff.lengthSquared() < PAN_DELTA) {
                    CCLOG("Match at %4.2f",diff.lengthSquared());
                    // Start the pan
                    _panid1 = it->first;
                    _panid2 = jt->first;
                    return;
                }
            }
        }
    }
}


#pragma mark -
#pragma mark Mouse Panning

/**
 * Called when a mouse button is initially pressed
 *
 * This is called in addition to the touch event, as mouse events are
 * all touch events. The mousepan boolean keeps them both from
 * being processed in future events.
 *
 * @param  event    the event storing the mouse state
 * @param  time     the time of the event
 */
void InputController::mouseDownCB(EventMouse* event, timestamp_t time) {
    if (_panid1 != -1) {
        return;
    }
    
    previous[-1] = event->getLocationInView();
    current[-1]  = event->getLocationInView();
    _mousepan = true;
}

/**
 * Called when a mouse button is released
 *
 * This is called in addition to the touch event, as mouse events are
 * all touch events. The mousepan boolean keeps them both from
 * being processed in future events.
 *
 * @param  event    the event storing the mouse state
 * @param  time     the time of the event
 */
void InputController::mouseUpCB(EventMouse* event, timestamp_t time) {
    if (_mousepan) {
        previous.erase(-1);
        current.erase(-1);
        _mousepan = false;
    }
}

/**
 * Called when the mouse moves
 *
 * This event registers ALL THE TIME. It does not care whether the mouse
 * button is held down or not. This makes this a very expensive listener
 *
 * @param  event    the event storing the mouse state
 * @param  time     the time of the event
 */
void InputController::mouseMoveCB(EventMouse* event, timestamp_t time) {
    if (_mousepan) {
        current[-1]  = event->getLocationInView();
    }
}
