//
//  PDInputController.h
//  PolyDemo
//
//  This input controller is primarily designed for touch and mouse control.  On mobile
//  you will notice that we need to use two fingers to pan.  However, when a mouse is
//  available, we can pan without that overhead.
//
//  Author: Walker White
//  Version: 2/10/16
//
#ifndef __PD_INPUT_CONTROLLER_H__
#define __PD_INPUT_CONTROLLER_H__
#include <unordered_map>
#include <cocos2d.h>
#include <cornell/CUTouchListener.h>
#include <cornell/CUMouseListener.h>

using namespace cocos2d;

#pragma mark Polled Input
/**
 * Class to represent multi-touch player input.
 *
 * Cocos2D really, REALLY wants to handle input via callbacks. This would make sense
 * for gesture recognition, except that Cocos2D has no built-in gesture recognition
 * (sigh).  So you have to implement manually, like we do here.  
 *
 * This input controller searches for a pan (e.g. a two-finger drag) and uses that to 
 * update an internal pan attribute.  This allows us to poll for panning, instead
 * of using a callback (which is preferable).  Honestly, I wish iOS had a polling
 * infrastructure like this.
 *
 * For multi-touch, there are two differently completely APIs that you can rely on. 
 * Depending on what you want to do, one is less laggy than another.  We have tried
 * them both, and one-by-one (instead of all-at-once) appears the be the better one.
 * Hence you should use TouchListener instead of MultiTouchListener.
 */
class InputController {
protected:
    // EVENT LISTENERS
    /** Listener to process touch events */
    TouchListener* _touchListener;
    /** Listener to process mouse events on a PC */
    MouseListener* _mouseListener;
    
    /** Whether or not this controller is currently active. */
    bool _active;
    /** The panning distance from the player input */
    Vec2 _pandelta;
    
    // TOUCH SUPPORT
    /** The first touch involved in active pan (or -1 if none) */
    int _panid1;
    /** The second touch involved in active pan (or -1 if none) */
    int _panid2;

    /** Location of the previous touches (organized by touch id) */
    std::unordered_map<int,Vec2> previous;
    /** Location of the most recent touches (organized by touch id) */
    std::unordered_map<int,Vec2> current;

    /** The initial touch location for the current gesture */
    Vec2 _dtouch;
    /** The timestamp for the beginning of the current gesture */
    timestamp_t _timestamp;

    /** Whether or not are in an active mouse pan */
    bool _mousepan;
    
    /** 
     * Search for a new pan
     *
     * This method searches for two fingers that are close together. The pan will
     * map to the two touch ids.  The pan will last as long as both fingers are 
     * down.  The fingers can be moved apart and the pan will still continue.
     * Only the first touch will determine the pan position.
     */
    void reacquirePan();
    
    
public:
    /**
     * Creates a new input controller.
     *
     * This constructor does NOT do any initialzation.  It simply allocates the
     * object. This makes it safe to use this class without a pointer.
     */
    InputController(); // Don't initialize.  Allow stack based
    
    /**
     * Disposes of this input controller, releasing all listeners.
     */
    ~InputController();
    
    /**
     * Initializes the input control for the given drawing scale.
     *
     * This method works like a proper constructor, initializing the input
     * controller and allocating memory.  However, it still does not activate
     * the listeners.  You must call start() do that.
     *
     * @return true if the controller was initialized successfully
     */
    bool init();
    
    /**
     * Starts the input processing for this input controller.
     *
     * This method must be called AFTER the input controller is initialized
     */
    void start();
    
    /**
     * Stops the input processing for this input controller.
     *
     * This method will not dispose of the input controller. It can be restarted.
     */
    void stop();
    
    /**
     * Processes the currently cached inputs.
     *
     * This method is used to to poll the current input state.  It will gather the
     * delta difference in the touches. Depending on the OS, we may see multiple updates
     * of the same touch in a single animation frame, so we need to accumulate all of
     * the data together.
     */
    void update(float dt);
    
    /**
     * Clears any buffered inputs so that we may start fresh.
     */
    void clear();

    
#pragma mark -
#pragma mark Input Results
/**
     * Returns the current pan delta.
     *
     * The delta is the amount that the user has moved a two-finger touch
     * since the last animation frame. This distance is measured according
     * to the first-finger touch.
     *
     * @return The input thrust
     */
    const Vec2& getPanDelta() { return _pandelta; }
    
    
#pragma mark -
#pragma mark Touch Callbacks
    
    /**
     * Callback for the beginning of a touch event
     *
     * @param t     The touch information
     * @param event The associated event
     *
     * @return True if the touch was processed; false otherwise.
     */
    bool    touchBeganCB(Touch* t, timestamp_t time);
    
    /**
     * Callback for the end of a touch event
     *
     * @param t     The touch information
     * @param event The associated event
     */
    void    touchEndedCB(Touch* t, timestamp_t time);
    
    /**
     * Callback for a touch movement event
     *
     * @param t     The touch information
     * @param event The associated event
     */
    void    touchMovedCB(Touch* t, timestamp_t time);
    
    /**
     * Callback for the cancellation of a touch event
     *
     * Cancellation occurs when an external event—for example, 
     * an incoming phone call—disrupts the current app’s event 
     * processing.
     *
     * @param t     The touch information
     * @param event The associated event
     */
    void    touchCancelCB(Touch* t, timestamp_t time);
    
    
#pragma mark -
#pragma mark Mouse Callbacks
    
    /**
     * Called when a mouse button is initially pressed
     *
     * This is called in addition to the touch event, as mouse events are
     * all touch events. The mousepan boolean keeps them both from
     * being processed in future events.
     *
     * @param  event    the event storing the mouse state
     * @param  time     the time of the event
     */
    void    mouseDownCB(EventMouse* event, timestamp_t time);
    
    /**
     * Called when a mouse button is released
     *
     * This is called in addition to the touch event, as mouse events are
     * all touch events. The mousepan boolean keeps them both from
     * being processed in future events.
     *
     * @param  event    the event storing the mouse state
     * @param  time     the time of the event
     */
    void    mouseUpCB(EventMouse* event, timestamp_t time);
    
    /**
     * Called when the mouse moves
     *
     * This event registers ALL THE TIME. It does not care whether the mouse
     * button is held down or not. This makes this a very expensive listener
     *
     * @param  event    the event storing the mouse state
     * @param  time     the time of the event
     */
    void    mouseMoveCB(EventMouse* event, timestamp_t time);
};

#endif /* defined(__PD_INPUT_CONTROLLER_H__) */